/***************************************************************************
                          touchtyper.cpp  -  description
                             -------------------
    begin                : tor jan  2 19:59:03 CET 2003
    copyright            : (C) 2003 by rn E. Hansen
    email                : orn.hansen@swipnet.se
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cassert>

#include <qlabel.h>
#include <qfile.h>
#include <qdir.h>
#include <qtextedit.h>
#include <qpushbutton.h>
#include <qpaintdevicemetrics.h>
#include <qdatetime.h>
#include <qtimer.h>
#include <qpainter.h>
#include <qprinter.h>
#include <qlcdnumber.h>
#include <qtextstream.h>

#include <kapp.h>
#include <kglobal.h>
#include <klocale.h>
#include <kfiledialog.h>
#include <kmessagebox.h>

#include "touchtyper.h"

//#include <X11/Xlib.h>

static int color_scheme[][13] = {
  { 1,1,2,2,3,4,4,5,5,6,7,8,8 },
  { 1,2,3,4,4,5,5,6,7,8,8,8,0 },
  { 1,2,3,4,4,5,5,6,7,8,8,8,0 },
  { 1,1,2,3,4,4,5,5,6,7,8,0,0 }
};

/**
  * The keysym table, provided here was used to get the keycode
  * number of the keyboards.  It's only here for reference, if
  * id ever come in handy again.
  */
static unsigned int key_syms[][13] = {
  { 0xb0,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x30,0xf6,0x2d },
  { 0x51,0x57,0x45,0x52,0x54,0x59,0x55,0x49,0x4f,0x50,0xf0,0x27,0    },
  { 0x41,0x53,0x44,0x46,0x47,0x48,0x4a,0x4b,0x4c,0xe6,0xfe51,0x2b,0  },
  { 0x3c,0x5a,0x58,0x43,0x56,0x42,0x4e,0x4d,0x2c,0x3a,0xfe,0,0       }
};

/**
  * These are the actual keycode of the keyboard part that is being
  * used in the representation dialog.  Note, that the codes used by
  * X11, are different from the codes used by Linux, which is also
  * different from the codes used by Qt.  ANNOYING!
  */
static unsigned int key_codes[][13] = {
  { 0x31,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f,0x10,0x11,0x12,0x13,0x14,0x15 },
  { 0x18,0x19,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,0x20,0x21,0x22,0x23,0    },
  { 0x26,0x27,0x28,0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,0x30,0x33,0    },
  { 0x5e,0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0   ,0    }
};

#define TT_TABLE_ROWS	4
#define TT_TABLE_COLS	13
#define TT_MAGIC_TEXT "#!TouchTyper_1.0"
#define TT_SEP_LEFT		'('
#define TT_SEP_RIGHT	')'
#define TT_SEP_VAL		';'

TouchTyper::TouchTyper(QWidget *parent, const char *name)
 : touchTrainer(parent, name)
{
  QObject *obj;
  QString s;
  QFile fil;

  secs_so_far   = 0;
  errors        = 0;
  keys_pressed  = 0;
  textGenerate  = false;
  capslock      = false;
  data          = "";
  printer       = new QPrinter;
  clock					= new QTime( 0,0,0,0 );
  timing				= new QTimer( this );
  displayKeysInState(0);
  mCourse.init( openFileByName("course.xml") );
	mExercize.advanceLevel( mCourse.nextLevel("") );
	mCourse.nextExercize( mExercize );
  markPosition();
  setFocusProxy( textDisplay );
  timingNum->display( "00:00" );
  textDisplay->installEventFilter( this );
  textDisplay->setCursorPosition( 0,0 );
	textDisplay->setTextFormat( Qt::PlainText );
  connect( timing,SIGNAL(timeout()),SLOT(slice()) );
}

TouchTyper::~TouchTyper()
{
}

bool
TouchTyper::getRowCol(const QChar ch_p,int &row_p,int &col_p, int &state_p)
{
	char ch;
	bool found = false;
  extern unsigned char getXKeysym(Display *,unsigned int,int);
	
	for( int i=0;i<(2*TT_TABLE_ROWS*TT_TABLE_COLS) && !found;i++ )
	{
		state_p = i/(TT_TABLE_COLS*TT_TABLE_ROWS);
		row_p		= (i/TT_TABLE_COLS)%TT_TABLE_ROWS;
		col_p		= i%TT_TABLE_COLS;
		if( key_codes[row_p][col_p] == 0 )
			continue;
		ch  = getXKeysym(x11Display(),key_codes[row_p][col_p],state_p);
		if( ch == ch_p.latin1() )
			found = true;
	}
	return found;
}

#include <qevent.h>

/** Shows which keys are pressed in the user interface. */
bool TouchTyper::eventFilter(QObject *o,QEvent *ep)
{
  QObject *obj;
  bool rv = true;
	
  if( ep->type() == QEvent::KeyRelease || ep->type() == QEvent::KeyPress )
  {
    bool keyState = (ep->type() == QEvent::KeyPress);
    QString objName = "";
    uchar kch = ((QKeyEvent*)ep)->ascii();
    uchar ch;
    int row,col,ks;
    ButtonState state;
    bool found=false;
    static int stateOfButtons = -1;

    switch( ((QKeyEvent*)ep)->key() )
    {
			case Qt::Key_Left:
				if( !keyState && textGenerate )
					textDisplay->moveCursor( QTextEdit::MoveBackward,false );
				break;
			case Qt::Key_Right:
				if( !keyState && textGenerate )
					textDisplay->moveCursor( QTextEdit::MoveForward,false );
				break;
			case Qt::Key_Up:
				if( !keyState && textGenerate )
					textDisplay->moveCursor( QTextEdit::MoveUp,false );
				break;
			case Qt::Key_Down:
				if( !keyState && textGenerate )
					textDisplay->moveCursor( QTextEdit::MoveDown,false );
				break;
      case Qt::Key_Backspace:
      case Qt::Key_Delete:
        setKeyState( "backspaceKey",keyState );
        if( !keyState && textGenerate )
          advancePos( '\b' );
        break;
      case Qt::Key_Meta:
      case Qt::Key_unknown:
      case Qt::Key_Alt:
        break;
      case Qt::Key_Shift:
        setKeyState( "lShiftKey",keyState );
        setKeyState( "rShiftKey",keyState );
        state = ((QKeyEvent *)ep)->state();
        found = !((state&QKeyEvent::ShiftButton) == QKeyEvent::ShiftButton);
        if( capslock )
          found = !found;
        displayKeysInState( found?1:0 );
        break;
      case Qt::Key_Enter:
      case Qt::Key_Return:
				// Temporary fix, to make sure Enter key is not the first typed character
				// if the person typed enter when selecting the file.
        if( !keyState && (keys_pressed || (textGenerate && textDisplay->text().length())))
          advancePos( '\n' );
        setKeyState( "returnKey",keyState );
        break;
      case Qt::Key_Space:
        if( !keyState )
          advancePos( ' ' );
        setKeyState( "spaceBarKey",keyState );
        break;
      case Qt::Key_Tab:
      case Qt::Key_Backtab:
        setKeyState( "tabKey",keyState );
        break;
      case Qt::Key_CapsLock:
        if( keyState )
          capslock = !capslock;
        setKeyState( "capsLockKey",capslock );
        displayKeysInState( capslock?1:0 );
        break;
      default:
        if( !keyState )
          advancePos( kch );
        if( !textGenerate && (!keyState && keys_pressed++ == 0) )
          timing->start( 1000,false );
				found = getRowCol(kch,row,col,ks);
        if( found )
          setKeyState( QString("row%1_%1").arg(row+1).arg(col+1),keyState );
        break;
    }
    return rv;
  }
  return QWidget::eventFilter( o,ep );
}

/** Alters the key state for a given key. */
void TouchTyper::setKeyState(const QString& keyName,bool set, const QString& keyText)
{
  QObject *obj;
  
  if( keyName != QString::null )
  {
    obj = QObject::child( keyName );
    if( obj && obj->inherits("QPushButton") ) {
      ((QPushButton*)obj)->setDown( set );
      if( keyText != QString::null )
        ((QPushButton*)obj)->setText( keyText );
    }
  }
}

/** No descriptions */
void TouchTyper::fileNew()
{
	QString s;
	
	if( !timing->isActive() && mExercize.accept(secs_so_far,errors ) )
	{
		KConfig *_cfg = KGlobal::config();
		
		mCourse.nextExercize( mExercize );
		if( mExercize.exercize() == QString::null )
		{
			mExercize.advanceLevel( mCourse.nextLevel(mExercize.level()) );
			mCourse.nextExercize( mExercize );
			if( mExercize.level() != QString::null )
			{
				KMessageBox::information(0,i18n("Congratulations, you have advanced to the level of ") + mExercize.level(),i18n("New Level"));
			} else {
				KMessageBox::information(0,i18n("Congratulations, you have graduated"),i18n("Graduation"));
			}
		}
		_cfg->setGroup("Exercise settings");
		_cfg->writeEntry("Level",mExercize.level());
		_cfg->writeEntry("Exercize",mExercize.exercizeNumber());
		_cfg->sync();
	}
	s = mExercize.level() + " - " + mExercize.exercizeNumber();
	labelLevel->setText( s );
	loadFileByName( openFileByName(mExercize.exercize().stripWhiteSpace()) );
}

/** Each second, this function is called to update
  * data. */
void TouchTyper::slice()
{
  QString s;

  s = clock->addSecs( ++secs_so_far ).toString();
  charsNum->display( QString("%1").arg(keys_pressed) );
  errorsNum->display( QString("%1").arg(errors) );
  timingNum->display( s.mid(3,5) );
}

/** No descriptions */
void TouchTyper::markPosition()
{
  int para,indx;

  textDisplay->getCursorPosition( &para,&indx );
  textDisplay->selectAll( false );
  textDisplay->setSelectionAttributes( 1,colorOf(textDisplay->text(para)[indx]),false );
  textDisplay->setSelection( para,indx,para,indx+1,1 );
}

/** No descriptions */
QColor TouchTyper::colorOf(const QChar& ch)
{
  int row,col,state;
  unsigned int xkc;

  if( ch.latin1() == ' ' )
    return QColor( 0,0,255 );
	if( getRowCol(ch,row,col,state) )
	{
    switch( color_scheme[row][col] )
    {
      case 1:
        return QColor( 0,170,0 );
      case 2:
        return QColor( 170,170,0 );
      case 3:
        return QColor( 170,85,0 );
      case 4:
        return QColor( 170,85,255 );
      case 5:
        return QColor( 255,170,0 );
      case 6:
        return QColor( 0,255,255 );
      case 7:
        return QColor( 255,85,0 );
      case 8:
        return QColor( 255,255,0 );
    }
  }
  return QColor( 0,0,0 );      
}

/** No descriptions */
void TouchTyper::advancePos(const QChar& ch)
{
  int paragraph,index;

  textDisplay->getCursorPosition( &paragraph,&index );
  if( textGenerate ) {
    switch( ch ) {
      case '\n': 
        textDisplay->doKeyboardAction( QTextEdit::ActionReturn );
        break;
      case '\b':
				textDisplay->doKeyboardAction( QTextEdit::ActionBackspace );
				break;
      default:
        if( uchar(ch.latin1()) >= ' ' ) {
					if( ch == '&' )
						textDisplay->insertAt( "&&",paragraph,index );
					else
	          textDisplay->insertAt( ch,paragraph,index );
					textDisplay->moveCursor( QTextEdit::MoveForward,false );
        }
    }
		return;
  }
  if( ch == '\n' ) {
    if( index < textDisplay->paragraphLength(paragraph) )
      errors += (textDisplay->paragraphLength(paragraph) - index);
		if( paragraph+1 >= textDisplay->paragraphs()-1 ) {
			timing->stop();
    } else {
			textDisplay->moveCursor( QTextEdit::MoveDown,false );
  	  textDisplay->moveCursor( QTextEdit::MoveLineStart,false );
		}
  } else if( uchar(ch.latin1()) >= ' ' ) {
    if( !textGenerate && textDisplay->text(paragraph)[index++] != ch )
      errors++;
    textDisplay->moveCursor( QTextEdit::MoveForward,false );
  }
  data += ch;
  markPosition();
}

/** Open the given file, which is a part of the program */
QString TouchTyper::openFileByName(const QString& fileName_p)
{
  QDir dirInfo(MYDATADIR);

  if( !dirInfo.exists(fileName_p) ) {
    dirInfo.setPath(".");
    if( !dirInfo.exists(fileName_p) )
      dirInfo.setPath("..");
  }
  return dirInfo.filePath(fileName_p);
}

/** No descriptions */
void TouchTyper::fileSave()
{
  QFile fil;
  
  if( fileName == "" || textGenerate == false )
    return;
  data = textDisplay->text();
	fil.setName( fileName );
  if( fil.open(IO_WriteOnly) )
  {
    QTextStream stream( &fil );

		if( textGenerate )
		{
			int row,col,state;
			
			stream << TT_MAGIC_TEXT << endl;
			for( int i=0;i<data.length();i++ )
			{
				if( data[i] == ' ' || data[i] == '\n' )
					stream << data[i];
				else if( getRowCol(data[i],row,col,state) )
					stream << TT_SEP_LEFT << row << TT_SEP_VAL << col << TT_SEP_VAL << state << TT_SEP_RIGHT;
				else
					stream << '!' << data[i];
			}
		}
    fil.close();
  }
}

/** No descriptions */
void TouchTyper::fileSaveAs()
{
  if( !textGenerate )
    return;
  fileName = KFileDialog::getSaveFileName(QString::null,QString::null,this,i18n("Save exercise to a file"));
  fileSave();
}

/** Display the keys on the keyboard, in state (Alt,Shift) */
void TouchTyper::displayKeysInState(int state)
{
  QString s,cstr;
  QChar ch;
  int row,col;
  extern unsigned char getXKeysym(Display *,unsigned int,int);

  for( int i=0;i<(TT_TABLE_ROWS*TT_TABLE_COLS);i++ )
  {
    row = i/TT_TABLE_COLS;
    col = i%TT_TABLE_COLS;
    if( key_codes[row][col] == 0 )
      continue;
    s  = QString("row%1_%2").arg(row+1).arg(col+1);
    ch = getXKeysym( x11Display(),key_codes[row][col],state );
		if( ch == '&' )
			cstr = "&&";
		else
			cstr = ch;
    setKeyState( s,false,cstr );
  }
}

/** Loads a test text, from an external file. */
void TouchTyper::loadFileByName(const QString& fileName_p)
{
  QString s;
	char buf[BUFSIZ];
  QFile fil(fileName_p);
  extern unsigned char getXKeysym(Display *,unsigned int,int);
  
	textGenerate = true;
  fileName = fileName_p;
  textDisplay->clear();
  if( fil.open(IO_ReadOnly) )
  {
    textGenerate = false;
		QTextStream ts(&fil);
		
		ts.readRawBytes( buf,strlen(TT_MAGIC_TEXT) );
		buf[strlen(TT_MAGIC_TEXT)] = 0;
		if( strncmp(buf,TT_MAGIC_TEXT,strlen(TT_MAGIC_TEXT)) == 0 )
		{
			bool ok = !ts.atEnd();
			int row,col,state;
			char ch,separator,terminator;
			
			s = "";
			textDisplay->clear();
			while( ok )
			{
				ts.readRawBytes( buf,1 );
				ch = buf[0];
				if( ch == TT_SEP_LEFT )
				{
					state = 0;
					ts >> row;
					ts >> separator;
					ts >> col;
					ts >> terminator;
					if( terminator == separator )
					{
						ts >> state;
						ts >> terminator;
					} else
						state = 0;
					ok = (separator == TT_SEP_VAL) && (terminator == TT_SEP_RIGHT);
					if( ok )
						s += getXKeysym( x11Display(),key_codes[row][col],state);
				} else {
					ok = (ch == ' ' || ch == '\n');
				  if ( ch == ' ' )
						s += ch;
				}
				if( ch == '\n' && s.length() )
				{
					textDisplay->insertParagraph( s,-1 );
					s = "";
				}
				ok = ok && !ts.atEnd();
			}
			if( s.stripWhiteSpace().length() )
				textDisplay->insertParagraph( s,-1 );
		}
    fil.close();
  }
	init();
}


/** Obtain copy of typed in text. */
QString TouchTyper::getData()
{
  return data;
}

/** What font to use, to represent data. */
QFont TouchTyper::dataFont()
{
  return textDisplay->font();
}


/*!
    \fn TouchTyper::exercizeLevel()
 */
QString TouchTyper::exercizeLevel()
{
    return mExercize.level();
}


/*!
    \fn TouchTyper::exercizeNumber()
 */
QString TouchTyper::exercizeNumber()
{
    return mExercize.exercizeNumber();
}


/*!
    \fn TouchTyper::setExercizeLevel(const QString&)
 */
void TouchTyper::setExercizeLevel(const QString& pLevel)
{
    mExercize.advanceLevel( pLevel );
}


/*!
    \fn TouchTyper::setExercizeNumber(const QString&)
 */
void TouchTyper::setExercizeNumber(const QString& pNumber)
{
	QDomElement elem = mCourse.getExercize(mExercize.level(),pNumber);
	
	if( elem.isNull() )
		mCourse.nextExercize( mExercize );
	else
		mExercize.init( elem );
}


/*!
    \fn TouchTyper::init()
 */
void TouchTyper::init()
{
  secs_so_far   = 0;
  keys_pressed  = 0;
  errors        = 0;
  data          = "";
  textDisplay->setCursorPosition( 0,0 );
  textDisplay->setReadOnly( !textGenerate );
	textDisplay->removeSelection( 1 );
	markPosition();
  timing->stop();
  charsNum->display( "0" );
  errorsNum->display( "0" );
  timingNum->display( "00:00" );
}
