/***************************************************************************
                          touchtyper.h  -  description
                             -------------------
    begin                : tor jan  2 19:59:03 CET 2003
    copyright            : (C) 2003 by rn E. Hansen
    email                : orn.hansen@swipnet.se
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TOUCHTYPER_H
#define TOUCHTYPER_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <kapp.h>
#include <qwidget.h>

#include "touchtrainer.h"
#include "trainingcourse.h"

/** TouchTyper is the base class of the project */
class TouchTyper : public touchTrainer
{
  Q_OBJECT 
  public:
    /** construtor */
    TouchTyper(QWidget* parent=0, const char *name=0);
    /** destructor */
    ~TouchTyper();
  /** What font to use, to represent data. */
  QFont dataFont();
  /** Obtain copy of typed in text. */
  QString getData();
  QString exercizeLevel();
  QString exercizeNumber();
  void setExercizeLevel(const QString&);
  void setExercizeNumber(const QString&);
protected: // Protected methods
  /** Shows which keys are pressed in the user interface. */
  bool eventFilter(QObject *,QEvent *);
protected: // Protected attributes
  /** Contains data on the key layout */
  QChar key_layout[2][4][13];
  /**  */
  QString data;
  /**  */
  bool capslock;
  /**  */
  int secs_so_far;
  /**  */
  int keys_pressed;
  /**  */
  QTimer* timing;
  /**  */
  QTime* clock;
  /**  */
  int errors;
  /**  */
  QPrinter* printer;
  /** mark weather the user wants to generate a test. */
	bool textGenerate;
  TrainingCourse mCourse;
  UserExercize mExercize;
protected slots: // Protected slots
  /** Alters the key state for a given key. */
  virtual void setKeyState(const QString&,bool,const QString& s=QString::null);
  /** No descriptions */
  void advancePos(const QChar&);
  /** No descriptions */
  QColor colorOf(const QChar&);
public slots: // Public slots
  void fileNew();
  /** Open the given file, which is a part of the program */
  QString openFileByName(const QString&);
  /** Loads a test text, from an external file. */
  virtual void loadFileByName(const QString&);
  /** No descriptions */
  void fileSaveAs();
  /** No descriptions */
  void fileSave();
protected slots: // Protected slots
  /** Each second, this function is called to update
data. */
  void slice();
public: // Public attributes
  /**  */
  QString fileName;
private: // Private methods
	/** Position of a character, inside our table */
	bool getRowCol(const QChar,int&, int&, int&);
  /** No descriptions */
  void markPosition();
  /** Display the keys on the keyboard, in state (Alt,Shift) */
  virtual void displayKeysInState(int);
    void init();
signals: // Signals
  /** Give an announcement, to be displayed on the statusbar. */
  void publicAnnouncement(const QString&);
};

#endif
