/***************************************************************************
                          touchtyperapp.cpp  -  description
                             -------------------
    begin                : sn jan 12 2003
    copyright            : (C) 2003 by rn E. Hansen
    email                : orn.hansen@swipnet.se
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// include files for QT
#include <qdir.h>
#include <qdatetime.h>
#include <qdatetimeedit.h>
#include <qpaintdevicemetrics.h>
#include <qlineedit.h>
#include <qradiobutton.h>
#include <qprinter.h>
#include <qpainter.h>
#include <qtextedit.h>

// include files for KDE
#include <kiconloader.h>
#include <kmessagebox.h>
#include <kfiledialog.h>
#include <kmenubar.h>
#include <kstatusbar.h>
#include <klocale.h>
#include <kconfig.h>
#include <kstdaction.h>

#include "touchtyper.h"
#include "touchtyperapp.h"

#define ID_STATUS_MSG 1

TouchTyperApp::TouchTyperApp(QWidget *,const char *name)
 : KMainWindow(0,name)
{
  config = kapp->config();

  initStatusBar();
  initActions();

  printer = new QPrinter;
  view = new TouchTyper(this,"trainer_window");
  connect( view,SIGNAL(publicAnnouncement(const QString&)),this,SLOT(slotStatusMsg(const QString&)) );
  setCentralWidget( view );
  setCaption( i18n("Typing exercise"),false );
  sizeForCentralWidgetSize( view->size() );

  readOptions();

//  fileOpenRecent->setEnabled( false );
//  fileOpen->setEnabled( false );
//  editCut->setEnabled(false);
//  editCopy->setEnabled(false);
//  editPaste->setEnabled(false);
}

TouchTyperApp::~TouchTyperApp()
{
}

/** Initialize the menubar actions. */
void TouchTyperApp::initActions()
{
//  fileNewWindow = new KAction(i18n("New &Window"), 0, 0, this, SLOT(slotFileNew()), actionCollection(),"file_new_window");
  fileNew = KStdAction::openNew(this, SLOT(slotFileNew()), actionCollection());
  fileOpen = KStdAction::open(this, SLOT(slotFileOpen()), actionCollection());
//  fileOpenRecent = KStdAction::openRecent(this, SLOT(slotFileOpenRecent(const KURL&)), actionCollection());
  fileSave = KStdAction::save(this, SLOT(slotFileSave()), actionCollection());
  fileSaveAs = KStdAction::saveAs(this, SLOT(slotFileSave()), actionCollection());
  fileClose = KStdAction::close(this, SLOT(slotFileQuit()), actionCollection());
  filePrint = KStdAction::print(this, SLOT(slotFilePrint()), actionCollection());
  fileQuit = KStdAction::quit(this, SLOT(slotFileQuit()), actionCollection());
//  editCut = KStdAction::cut(this, SLOT(slotEditCut()), actionCollection());
//  editCopy = KStdAction::copy(this, SLOT(slotEditCopy()), actionCollection());
//  editPaste = KStdAction::paste(this, SLOT(slotEditPaste()), actionCollection());
//  editProperties = new KAction(i18n("&Properties"),0,0,this,SLOT(slotEditProperties()),actionCollection(),"properties");
  viewToolBar = KStdAction::showToolbar(this, SLOT(slotViewToolBar()), actionCollection());
  viewStatusBar = KStdAction::showStatusbar(this, SLOT(slotViewStatusBar()), actionCollection());

//  fileNewWindow->setStatusText(i18n("Opens a new application window"));
  fileNew->setStatusText(i18n("Starts a new exercise"));
  fileOpen->setStatusText(i18n("Opens the resulting text, in a view window."));
//  fileOpenRecent->setStatusText(i18n("Opens a recently taken exercise results."));
  fileSave->setStatusText(i18n("Saves the current exercise results"));
  fileSaveAs->setStatusText(i18n("Saves the current exercise results as..."));
  fileClose->setStatusText(i18n("Closes the actual document"));
  filePrint ->setStatusText(i18n("Prints out the exercise results"));
  fileClose->setStatusText(i18n("Quits the application"));
//  editCut->setStatusText(i18n("Cuts the selected section and puts it to the clipboard"));
//  editCopy->setStatusText(i18n("Copies the selected section to the clipboard"));
//  editPaste->setStatusText(i18n("Pastes the clipboard contents to actual position"));
//  editProperties->setStatusText(i18n("Modified exercise properties"));
  viewToolBar->setStatusText(i18n("Enables/disables the toolbar"));
  viewStatusBar->setStatusText(i18n("Enables/disables the statusbar"));

  createGUI();
}

/** Initialize status bar layout. */
void TouchTyperApp::initStatusBar()
{
  ///////////////////////////////////////////////////////////////////
  // STATUSBAR
  // TODO: add your own items you need for displaying current application status.
  statusBar()->insertItem(i18n("Ready."), ID_STATUS_MSG);
}

/** No descriptions */
void TouchTyperApp::saveOptions()
{
  config->setGroup("General Options");
  config->writeEntry("Show Toolbar", viewToolBar->isChecked());
  config->writeEntry("Show Statusbar",viewStatusBar->isChecked());
  config->writeEntry("ToolBarPos", (int) toolBar("mainToolBar")->barPos());
	
	saveProperties( config );
}

/** No descriptions */
void TouchTyperApp::readOptions()
{
  config->setGroup("General Options");

  // bar status settings
  bool bViewToolbar = config->readBoolEntry("Show Toolbar", true);
  viewToolBar->setChecked(bViewToolbar);
  slotViewToolBar();

  bool bViewStatusbar = config->readBoolEntry("Show Statusbar", true);
  viewStatusBar->setChecked(bViewStatusbar);
  slotViewStatusBar();


  // bar position settings
  KToolBar::BarPosition toolBarPos;
  toolBarPos=(KToolBar::BarPosition) config->readNumEntry("ToolBarPos", KToolBar::Top);
  toolBar("mainToolBar")->setBarPos(toolBarPos);

	readProperties( config );
}

/** Save property settings, for the current exercise. */
void TouchTyperApp::saveProperties(KConfig *_cfg)
{
	qDebug("saving ...");
	_cfg->setGroup("Exercise settings");
	_cfg->writeEntry("Level",view->exercizeLevel());
	_cfg->writeEntry("Exercize",view->exercizeNumber());
}

/** Read properties, for application. */
void TouchTyperApp::readProperties(KConfig *_cfg)
{
	QString s;

	qDebug("reading ...");	
	_cfg->setGroup("Exercise settings");
	s = _cfg->readEntry("Level","beginner");
	view->setExercizeLevel( s );
	s = _cfg->readEntry("Exercize","0");
	view->setExercizeNumber( s );
}

/** Start a new exercise */
void TouchTyperApp::slotFileNew()
{
  slotStatusMsg(i18n("Starting a new exercise"));
  view->fileNew();
  slotStatusMsg(i18n("Ready."));
}

/** Save the current exercise data. */
void TouchTyperApp::slotFileSave()
{
  slotStatusMsg(i18n("Saving data..."));
  view->fileSave();
  slotStatusMsg(i18n("Ready."));
}

/** Save current exercise data, under a new filename. */
void TouchTyperApp::slotFileSaveAs()
{
  slotStatusMsg(i18n("Saving data as..."));
  view->fileSaveAs();
  slotStatusMsg(i18n("Ready."));
}

/** Print the current exercise data. */
void TouchTyperApp::slotFilePrint()
{
  const int margin=10;
  int pageNo = 1;
  int YPos = 0;
  int indx = 0;
  int lengd;
  QString s;

  slotStatusMsg(i18n("Printing..."));
  s = view->getData();
  if( printer->setup(this) )
  {
    QPainter p;
    if( !p.begin(printer) )
      return;
    p.setFont( view->dataFont() );
    QFontMetrics fm = p.fontMetrics();
    QPaintDeviceMetrics metrics( printer );
//    if ( view->printHeader(p,metrics.width(),fm.lineSpacing(),margin) )
//      YPos += 3*fm.lineSpacing();
    while( s.length() > indx+1 )
    {
      if( margin + YPos > metrics.height() - 2*margin )
      {
        printFooter(p,metrics.width(),metrics.height()-margin,fm.lineSpacing(),++pageNo);
        printer->newPage();
        YPos = 0;
//        if( view->printHeader(p,metrics.width(),fm.lineSpacing(),margin) )
//          YPos += 3*fm.lineSpacing();
      }
      for(lengd=1;s.length()>indx+lengd && fm.width(s.mid(indx,lengd))<metrics.width()-(2*margin);lengd++)
        if( s[indx+lengd-1] == '\n' )
          break;
      lengd--;
      if( s.length() > indx+lengd+1 )
        if( !s[indx+lengd].isSpace() && s.mid(indx,lengd).contains(' ') )
        {
          do {
            lengd--;
          } while( !s[indx+lengd].isSpace() && lengd > 1);
        }
      p.drawText( margin,margin+YPos,metrics.width(),fm.lineSpacing(),ExpandTabs|DontClip,s.mid(indx,lengd));
      YPos += fm.lineSpacing();
      indx += lengd;
      while( s[indx].isSpace() && s.length() > indx )
      {
        if( s[indx] == '\n' )
          YPos += fm.lineSpacing();
        indx++;
      }
    }
    printFooter(p,metrics.width(),metrics.height()-margin,fm.lineSpacing(),pageNo);
    p.end();
  }
  slotStatusMsg(i18n("Ready."));
}

/** Quit the application. */
void TouchTyperApp::slotFileQuit()
{
  close();
}

/** No descriptions */
void TouchTyperApp::slotViewToolBar()
{
  slotStatusMsg(i18n("Toggling toolbar..."));
  ///////////////////////////////////////////////////////////////////
  // turn Toolbar on or off
  if(!viewToolBar->isChecked())
  {
    toolBar("mainToolBar")->hide();
  }
  else
  {
    toolBar("mainToolBar")->show();
  }

  slotStatusMsg(i18n("Ready."));
}

/** No descriptions */
void TouchTyperApp::slotViewStatusBar()
{
  slotStatusMsg(i18n("Toggle the statusbar..."));
  ///////////////////////////////////////////////////////////////////
  //turn Statusbar on or off
  if(!viewStatusBar->isChecked())
  {
    statusBar()->hide();
  }
  else
  {
    statusBar()->show();
  }

  slotStatusMsg(i18n("Ready."));
}

/** No descriptions */
void TouchTyperApp::slotStatusMsg(const QString& text)
{
  statusBar()->clear();
  statusBar()->changeItem(text, ID_STATUS_MSG);
}

/** Query if exercise can be closed. */
bool TouchTyperApp::queryClose()
{
  return true;
}

/** Query if application may exit. */
bool TouchTyperApp::queryExit()
{
  return queryClose();
}

/** Open up a new exercise file, to type after. */
void TouchTyperApp::slotFileOpen()
{
  view->loadFileByName( KFileDialog::getOpenFileName(QString::null,"*.test",this,i18n("Open a test file")) );
}

/** Print a footer on the page. */
void TouchTyperApp::printFooter(QPainter& p,int width,int height,int spacing,int page)
{
  int x;
  QString s;

  s = QString(i18n("Page %1")).arg(page);
  x = (width - p.fontMetrics().width(s))/2;
  p.drawText(x,height-spacing,width,spacing,ExpandTabs|DontClip,s);
}
